package checkers.source;

import java.util.*;

/**
 * Represents the outcome of a typechecking operation. {@link Result}s created
 * during typechecking can be reported using {@link SourceChecker#report}, which
 * ultimately delivers an error or warning message via the JSR 199 compiler
 * interface.
 *
 * @see SourceChecker#report
 */
public final class Result {

    private static enum Type {
        SUCCESS, FAILURE, WARNING;

        /**
         * @return whichever of the given types is most serious
         */
        public static final Type merge(Type a, Type b) {
            if (a == FAILURE || b == FAILURE)
                return FAILURE;
            else if (a == WARNING || b == WARNING)
                return WARNING;
            else
                return SUCCESS;
        }
    };

    /** The type of result (success, warning, failure). */
    private final Type type;

    /** The messages for this result. */
    private final List<String> messages;

    /** The success result. */
    public static final Result SUCCESS = new Result(Type.SUCCESS, null);

    /**
     * Creates a new failure result with the given message.
     * 
     * @param message
     *            the reason for failure
     * @return the failure result
     */
    public static Result failure(/* @NonNull */String message) {
        return new Result(Type.FAILURE, Collections.singletonList(message));
    }

    /**
     * Creates a new warning result with the given message.
     * 
     * @param message
     *            the reason for the warning
     * @return the warning result
     */
    public static Result warning(/* @NonNull */String message) {
        return new Result(Type.WARNING, Collections.singletonList(message));
    }

    /**
     * Creates a non-success result.
     * 
     * @param type
     *            the type of result
     * @param message
     *            the message for this result
     */
    private Result(Type type, List<String> messages) {
        this.type = type;
        this.messages = new LinkedList<String>();
        if (messages != null)
            this.messages.addAll(messages);
    }

    /**
     * Merges two results into one.
     * 
     * @param r
     *            the result to merge with this result
     * @return a result that is the success result if both this and {@code r}
     *         are success results, or a result that has the more significant
     *         type (failure > warning > success) and the messages of both this
     *         result and {@code r}
     */
    public Result merge(Result r) {
        if (r == null)
            return this;

        if (r.isSuccess() && this.isSuccess())
            return SUCCESS;

        List<String> messages = new LinkedList<String>();
        messages.addAll(this.messages);
        messages.addAll(r.messages);
        return new Result(Type.merge(r.type, this.type), messages); 
                
    }

    /**
     * @return true if the result is not a failure or warning
     */
    public boolean isSuccess() {
        return type == Type.SUCCESS;
    }

    /**
     * @return true if the result is a failure
     */
    public boolean isFailure() {
        return type == Type.FAILURE;
    }

    /**
     * @return true if the result is a warning
     */
    public boolean isWarning() {
        return type == Type.WARNING;
    }

    /**
     * @return the message associated with the result
     */
    public String[] getMessages() {
        return messages.toArray(new String[messages.size()]);
    }

    @Override
    public String toString() {
        switch (type) {
        case FAILURE:
            return "FAILURE: " + messages;
        case WARNING:
            return "WARNING: " + messages;
        case SUCCESS:
        default:
            return "SUCCESS";
        }
    }
}
