package checkers.source;

import java.io.IOException;
import java.util.*;

import javax.annotation.processing.*;
import javax.lang.model.element.*;
import javax.tools.Diagnostic;

import checkers.types.*;

import com.sun.source.tree.*;
import com.sun.source.util.Trees;

/**
 * An abstract annotation processor designed for conveniently implementing a
 * source-file checker for a JSR 308 type-checker plugin. It provides an
 * interface to {@code javac}'s annotation processing API, routines for error
 * reporting via the JSR 199 compiler API, and an implementation for using a
 * {@link SourceVisitor} to perform the type-checking.
 * 
 * <p>
 * 
 * Subclasses should provide the necessary overrides for configuring an
 * {@link AbstractProcessor}:
 * <ul>
 *  <li>{@link AbstractProcessor#getSupportedAnnotationTypes}
 *  <li>{@link AbstractProcessor#getSupportedSourceVersion}
 * </ul>
 *
 * <p>
 *  
 * In addition, the following methods allow subclasses to customize the default
 * implementation provided by this class:
 * <ul>
 *  <li>{@link SourceChecker#getMessages} (for type-qualifier specific error messages)
 *  <li>{@link SourceChecker#getSourceVisitor} (for a custom {@link SourceVisitor})
 *  <li>{@link SourceChecker#getFactory} (for a custom {@link AnnotatedTypeFactory})
 * </ul>
 */
public abstract class SourceChecker extends AbstractProcessor {

    /** Provides access to compiler helpers/internals. */
    protected ProcessingEnvironment env;

    /** Used for scanning a source tree. */
    protected SourceVisitor visitor;

    /** Maps error keys to localized/custom error messages. */
    protected Properties messages;

    /** Used to report error messages and warnings via the compiler. */
    protected Messager messager;

    /** Used as a helper to the {@link SourceVisitor}. */
    protected Trees trees;

    /** The source tree that's being scanned. */
    protected CompilationUnitTree currentRoot;

    /**
     * @return the {@link ProcessingEnvironment} that was supplied to this
     *         checker
     */
    public ProcessingEnvironment getProcessingEnvironment() {
        return this.env;
    }

    /**
     * @param env the {@link ProcessingEnvironment} for the factory
     * @param root the AST root for the factory
     * @return an {@link AnnotatedTypeFactory} for use by typecheckers
     */
    public AnnotatedTypeFactory getFactory(ProcessingEnvironment env, CompilationUnitTree root) {
        return new AnnotatedTypeFactory(env, root);
    }
    
    /**
     * Provides the {@link SourceVisitor} that the checker should use to scan
     * input source trees.
     * 
     * @param root
     *            the AST root
     * @return a {@link SourceVisitor} to use to scan source trees
     */
    protected abstract SourceVisitor getSourceVisitor(CompilationUnitTree root);

    /**
     * Provides a mapping of error keys to localized or custom error messages.
     * 
     * @param fileName the name of the messages file
     * @throws IOException
     *             if the messages couldn't be loaded
     * @return a {@link Properties} that maps error keys to error message text
     */
    protected abstract Properties getMessages(String fileName) throws IOException;

    /**
     * {@inheritDoc}
     * 
     * @see javax.annotation.processing.AbstractProcessor#init(javax.annotation.processing.ProcessingEnvironment)
     */
    @Override
    public synchronized void init(ProcessingEnvironment processingEnv) {
        super.init(processingEnv);
        this.env = processingEnv;

        // Grab the Trees and Messager instances now; other utilities
        // (like Types and Elements) can be retrieved by subclasses.
        this.trees = Trees.instance(processingEnv);
        this.messager = processingEnv.getMessager();

        try {
            this.messages = getMessages("checker.properties");
        } catch (IOException e) {
            System.err.println("messages not found: " + e.getMessage());
            this.messages = new Properties();
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @see javax.annotation.processing.AbstractProcessor#process(java.util.Set,
     *      javax.annotation.processing.RoundEnvironment)
     */
    @Override
    public boolean process(Set<? extends TypeElement> annotations,
            RoundEnvironment roundEnv) {

        Set<? extends Element> elements = roundEnv.getRootElements();
        assert elements != null;

        if (elements.isEmpty())
            return false;

        for (Element e : elements) {

            if (trees.getPath(e) == null)
                continue;

            currentRoot = trees.getPath(e).getCompilationUnit();
            visitor = getSourceVisitor(currentRoot);
            visitor.scan(trees.getTree(e), null);
        }

        return true;
    }

    /**
     * Prints an message (error, warning, note, etc.) via JSR-269.
     * 
     * @param kind
     *            the type of message to print
     * @param source
     *            the object from which to obtain source position information
     * @param msg
     *            the message key to print
     * @param args
     *            arguments for interpolation in the message key
     * @throws IllegalArgumentException
     *             if {@code source} is neither a {@link Tree} nor an
     *             {@link Element}
     * 
     * @see Diagnostic
     */
    protected void message(Diagnostic.Kind kind, Object source, Object msg,
            Object... args) {

        assert messages != null : "null messages";

        if (kind == Diagnostic.Kind.NOTE) {
            System.err.println("(NOTE) " + String.format(msg.toString(), args));
            return;
        }

        String message = messages.getProperty(msg.toString(), "("
                + msg.toString() + ")");
        String messageText = String.format(message, args);

        if (source instanceof Element)
            messager.printMessage(kind, messageText, (Element) source);
        else if (source instanceof Tree)
            messager.printMessageFromTree(kind, messageText, (Tree) source,
                    currentRoot);
        else
            throw new IllegalArgumentException("invalid position source: "
                    + source.getClass().getName());
    }

    /**
     * Reports a result. By default, it prints it to the screen via the
     * compiler's internal messager if the result is non-success; otherewise,
     * the method returns with no side-effects.
     * 
     * @param r
     *            the result to report
     * @param src
     *            the position object associated with the result
     */
    public void report(final Result r, final Object src) {

        if (r.isSuccess())
            return;

        for (String msg : r.getMessages()) {
            if (r.isFailure())
                this.message(Diagnostic.Kind.ERROR, src, msg);
            else if (r.isWarning())
                this.message(Diagnostic.Kind.MANDATORY_WARNING, src, msg);
            else
                this.message(Diagnostic.Kind.NOTE, src, msg);
        }
    }
}
