package checkers.subtype;

import java.io.*;
import java.lang.annotation.Annotation;
import java.util.*;

import javax.lang.model.SourceVersion;
import javax.lang.model.element.*;
import javax.lang.model.util.Elements;

import checkers.source.*;
import checkers.types.*;

import com.sun.source.tree.CompilationUnitTree;

/**
 * An annotation processor for typechecking qualifiers for which the qualified
 * type is the subtype of the unqualified type.
 */
public class SubtypeChecker extends SourceChecker {

    /** The annotation to check. */
    protected Class<? extends Annotation> annotation;

    /**
     * Creates an annotation processor for checking subtype qualifiers.
     */
    public SubtypeChecker() {
        this.annotation = getCheckedAnnotation();
    }

    /**
     * Gets the class of the annotation which will be checked by this annotation
     * processor.
     * 
     * @return the class of the type qualifier annotation
     */
    @SuppressWarnings("unchecked")
    protected Class<? extends Annotation> getCheckedAnnotation() {
        
        String annotation = System.getProperty("annotation");
        
        try {
            return (Class<? extends Annotation>)Class.forName(annotation);
        } catch (Exception e) {
            return Annotation.class;
        }        
    }
    
    @Override
    public Set<String> getSupportedAnnotationTypes() {
        return Collections.singleton("*");
    }

    @Override
    public SourceVersion getSupportedSourceVersion() {
        return SourceVersion.RELEASE_6;
    }

    @Override
    protected/* @NonNull */Properties getMessages(String fileName) throws IOException {
        
        String path = System.getProperty("messages", "."); 
        Properties p = new Properties();
        p.load(new FileInputStream(path + File.separator + fileName));
        return p;
    }

    /**
     * Determines whether the first type is a subtype of the second type with
     * respect to the annotations on that type.
     *
     * @param type the type which may or may not be a subtype
     * @param other the parent type
     * @return true if {@code type} is a {@code subtype} of other with respect
     *         to its type qualifier annotations, false otherwise
     */
    public boolean isSubtype(AnnotatedClassType type, AnnotatedClassType other) {
        if (!type.isAnnotatedWith(this.annotation) && 
                other.isAnnotatedWith(this.annotation))
            return false;
        return true;
    }
    
    /**
     * Determines whether assignment is permissible from one type to another
     * given the rules of the type qualifier annotation being checked.
     *
     * @param from the right-hand type
     * @param to the left-hand type
     * @return true if {@code to} can be assigned to {@code from} with respect
     *         to the type qualifier annotation being checked
     */
    public boolean isAssignable(AnnotatedClassType from, AnnotatedClassType to) {
        
//        System.out.println("FROM: " + from);
//        System.out.println("TO: " + to);

        if (from.getElement() != null) {
            Elements elts = env.getElementUtils();
            Element fromElt = from.getElement();
            if (elts.getPackageOf(fromElt).getQualifiedName().toString().startsWith("java"))
                return true;
        }
        
        if (!isSubtype(from, to))
            return false;

        boolean invalid = false;

        for (AnnotationLocation location : to.getAnnotatedLocations()) {
            if (!from.hasAnnotationAt(this.annotation, location)
                    && !from.hasWildcardAt(location)) {
                invalid = true;
                break;
            }
        }

        for (AnnotationLocation location : from.getAnnotatedLocations()) {
            if (!to.hasAnnotationAt(this.annotation, location)
                    && !to.hasWildcardAt(location)) {
                invalid = true;
                break;
            }
        }

        if (invalid)
            return false;
        
        return true;        
    }

    @Override
    protected SourceVisitor getSourceVisitor(CompilationUnitTree root) {
        return new SubtypeVisitor(this, root);
    }
}
