package checkers.subtype;

import java.util.*;

import javax.lang.model.element.*;
import javax.lang.model.type.TypeMirror;

import checkers.source.*;
import checkers.types.*;

import com.sun.source.tree.*;
import com.sun.source.util.TreePath;

/**
 * An type-checking visitor for type qualifiers for which the qualified type
 * is the subtype of the unqualified type.
 */
public class SubtypeVisitor extends SourceVisitor {

    /** Whether or not system classes should not be checked against. */
    protected final boolean SKIP_SYSTEM_CLASSES = 
            Boolean.valueOf(System.getProperty("checkers.skipSystemClasses", "true"));

    /** The checker associated with this visitor. */
    protected final SubtypeChecker checker;

    /** 
     * Creates a visitor for checking subtype qualifiers, using the given 
     * checker, which will operate on the given syntax tree root.
     * 
     * @param checker the checker associated with this visitor
     * @param root the root of the syntax tree to check
     */
    public SubtypeVisitor(SubtypeChecker checker, CompilationUnitTree root) {
        super(checker, root);
        this.checker = checker;
    }

    /**
     * @see com.sun.source.util.TreeScanner#visitAssignment(com.sun.source.tree.AssignmentTree,
     *      java.lang.Object)
     */
    @Override
    public Void visitAssignment(AssignmentTree node, Void p) {

        AnnotatedClassType expression = factory.getClass(node.getExpression());
        AnnotatedClassType variable = factory.getClass(node.getVariable());

        if (!checker.isAssignable(expression, variable)) {
            checker.report(Result.failure("assignment.invalid"), node
                    .getExpression());
        }

        return super.visitAssignment(node, p);
    }

    /**
     * @see com.sun.source.util.TreeScanner#visitVariable(com.sun.source.tree.VariableTree,
     *      java.lang.Object)
     */
    @Override
    public Void visitVariable(VariableTree node, Void p) {

        if (node.getInitializer() == null)
            return super.visitVariable(node, p);

        AnnotatedClassType init = factory.getClass(node.getInitializer());
        AnnotatedClassType variable = factory.getClass(node);

        if (!checker.isAssignable(init, variable))
            checker.report(Result.failure("assignment.invalid"), variable
                    .getElement());

        return super.visitVariable(node, p);
    }

    /**
     * @see com.sun.source.util.TreeScanner#visitMethodInvocation(com.sun.source.tree.MethodInvocationTree,
     *      java.lang.Object)
     */
    @Override
    public Void visitMethodInvocation(MethodInvocationTree node, Void p) {

        AnnotatedMethodType method = factory.getMethod(node);
        List<AnnotatedClassType> parameters = method
                .getAnnotatedParameterTypes();

        // Skip checking method invocations (i.e., super) inside the
        // (synthetic) constructors of anonymous inner classes, since they're
        // synthesized without any annotations on their arguments.
        TreePath path = TreePath.getPath(root, node);
        if (InternalUtils.isAnonymousConstructor(path))
            return super.visitMethodInvocation(node, p);

        List<AnnotatedClassType> arguments = new LinkedList<AnnotatedClassType>();
        for (ExpressionTree arg : node.getArguments())
            arguments.add(factory.getClass(arg));

        for (int i = 0; i < arguments.size(); i++) {
            if (!checker.isSubtype(arguments.get(i), parameters.get(i)))
                checker.report(Result.failure("argument.invalid"), node);
        }

        if (node.getMethodSelect() instanceof MemberSelectTree) {
            MemberSelectTree ms = (MemberSelectTree) node.getMethodSelect();
            AnnotatedClassType receiver = factory.getClass(ms.getExpression());

            Element methodElt = method.getAnnotatedReceiverType().getElement();
            if (SKIP_SYSTEM_CLASSES
                    && elements.getPackageOf(methodElt).getQualifiedName()
                            .toString().startsWith("java"))
                return super.visitMethodInvocation(node, p);

            // TODO *** probably shouldn't be here?
            ExpressionTree expr = ms.getExpression();
            if (expr != null) {
                Element exprElt = InternalUtils.symbol(expr);
                if (exprElt != null && exprElt.getSimpleName().contentEquals("this"))
                    return super.visitMethodInvocation(node, p);
                    
            }
            
            if (!checker.isSubtype(receiver, method.getAnnotatedReceiverType()))
                checker.report(Result.failure("receiver.invalid"), ms);
        }

        return super.visitMethodInvocation(node, p);

    }

    /**
     * @see com.sun.source.util.TreeScanner#visitReturn(com.sun.source.tree.ReturnTree,
     *      java.lang.Object)
     */
    @Override
    public Void visitReturn(ReturnTree node, Void p) {

        if (node.getExpression() == null)
            return super.visitReturn(node, p);

        AnnotatedClassType ret = factory.getClass(node.getExpression());
        Tree enclosingMethod = InternalUtils.enclosingMethod(TreePath.getPath(
                root, node));

        AnnotatedMethodType meth = factory.getMethod(enclosingMethod);
        
        if (!checker.isAssignable(ret, meth.getAnnotatedReturnType()))
            checker.report(Result.failure("return.invalid"), node);

        return super.visitReturn(node, p);
    }

    private void closureMethods(TypeMirror type,
            Map<Element, List<? extends Element>> methods) {
        for (TypeMirror t : types.directSupertypes(type)) {
            Element te = types.asElement(t);
            methods.put(te, te.getEnclosedElements());
            closureMethods(t, methods);
        }
    }

    @Override
    public Void visitMethod(MethodTree node, Void p) {

        ExecutableElement elt = (ExecutableElement) InternalUtils.symbol(node);

        AnnotatedMethodType method = factory.getMethod(node);

        Element encl = elt.getEnclosingElement();
        if (encl.getKind() != ElementKind.CLASS)
            return super.visitMethod(node, p);

        TypeElement te = (TypeElement) encl;

        Map<Element, List<? extends Element>> methods = new LinkedHashMap<Element, List<? extends Element>>();

        closureMethods(te.asType(), methods);

        ExecutableElement overridingMethod = null;
        outer: for (Element superElt : methods.keySet()) {
            for (Element member : superElt.getEnclosedElements()) {
                if (member.getKind() != ElementKind.METHOD)
                    continue;

                ExecutableElement meth = (ExecutableElement) member;
                if (elements.overrides(elt, meth, (TypeElement) superElt)) {
                    overridingMethod = meth;
                    break outer;
                }
            }
        }

        // TODO *** re-enable for method override checking

//        if (overridingMethod != null) {
//            AnnotatedMethodType superMethod =
//                factory.getMethod(overridingMethod);
//
//            if (!checker.isAssignable(method.getAnnotatedReturnType(),
//                    superMethod.getAnnotatedReturnType()))
//                checker.report(Result.failure("override.return.invalid"), node);
//
//            List<AnnotatedClassType> methodParams =
//                method.getAnnotatedParameterTypes();
//            List<AnnotatedClassType> superMethodParams =
//                superMethod.getAnnotatedParameterTypes();
//            assert methodParams.size() == superMethodParams.size();
//
//            for (int i = 0; i < superMethodParams.size(); i++) {
//                if (!checker.isAssignable(superMethodParams.get(i),
//                        methodParams.get(i)))
//                    checker.report(Result.failure("override.param.invalid"), node);
//            }
//        }

        return super.visitMethod(node, p);
    }
}
