package checkers.types;

import java.lang.annotation.Annotation;
import java.util.*;

import javax.lang.model.element.Element;
import javax.lang.model.type.*;

import com.sun.source.tree.Tree;

/**
 * The base class for representing annotations on types. It provides the means
 * to obtain annotations written on the declarations of a type from any valid
 * use of a type, and it presents annotations in a uniform way, regardless of
 * whether the annotations were written directly on the type, inherited from the
 * supertype, or specified via a "default" annotation.
 */
public class AnnotatedClassType {

    /** The element node related to this annotated type. */
    protected Element element;

    /** The tree node related to this annotated type. */
    protected Tree tree;

    /** The set of annotations present on the outer type. */
    protected final Set<Class<? extends Annotation>> outer;

    /** The set of annotations present on the inner type, if any. */
    protected final Map<AnnotationLocation, Set<Class<? extends Annotation>>> inner;

    /** The collection of indirectly present annotation types. */
    protected final Set<Class<? extends Annotation>> includes;

    /** The collection of annotation types that should be ignored if present. */
    protected final Set<Class<? extends Annotation>> excludes;

    /**
     * Creates an annotated class type. This class type will be related to the
     * given tree node, making it appropriate for annotated types corresponding
     * directly to most program elements (i.e., fields, variables, method
     * invocations).
     */
    protected AnnotatedClassType() {

        // Initialize annotation storage.
        this.outer = new HashSet<Class<? extends Annotation>>();
        this.inner = new HashMap<AnnotationLocation, Set<Class<? extends Annotation>>>();
        this.includes = new HashSet<Class<? extends Annotation>>();
        this.excludes = new HashSet<Class<? extends Annotation>>();
    }

    /**
     * Determines whether the annotated type contains an annotation (either
     * directly or indirectly) with the given class (on the outer type only).
     * 
     * @param annotation
     *            the class token for the annotation to check for
     * @return true if the given annotation is present on this annotated type,
     *         false otherwise
     */
    public boolean isAnnotatedWith(Class<? extends Annotation> annotation) {

        // If this type explicitly excludes this annotation type, regard it
        // as !isAnnotatedWith no matter what.
        if (excludes.contains(annotation))
            return false;

        if (includes.contains(annotation))
            return true;

        if (outer.contains(annotation))
            return true;

        return false;
    }

    /**
     * Determines whether the annotated type has an inner type annotation with
     * the given class at the given location.
     * 
     * @param annotation
     *            the annotation to check for
     * @param location
     *            the location to check
     * @return true if the given annotation is present at the given location
     *         false otherwise
     */
    public boolean hasAnnotationAt(Class<? extends Annotation> annotation,
            AnnotationLocation location) {

        if (inner.containsKey(location)
                && inner.get(location).contains(annotation))
            return true;

        return false;
    }

    /**
     * Determines whether the type backing this annotated type has a wildcard
     * ("?") at location specified by the given integer list.
     * 
     * @param location
     *            a list of integers specifying the location within an inner
     *            type; the <i>i</i>th integer in the list holds the index
     *            (from zero) of a parameter at level <i>i</i> in the type
     *            hierarchy
     * @return true if there is a wildcard at the given location, false
     *         otherwise
     */
    public boolean hasWildcardAt(AnnotationLocation location) {

        if (element == null)
            return false;
        
        TypeMirror type = element.asType();

        if (!(type instanceof DeclaredType))
            return false;

        Queue<Integer> queue = new LinkedList<Integer>();
        queue.addAll(location.asList());

        // Walk down the type hierarchy according to the integer list,
        // locating the parameter that the list specifies.

        TypeMirror current = type;
        while (!queue.isEmpty()) {
            if (current instanceof DeclaredType) {
                DeclaredType declared = (DeclaredType) current;
                int index = queue.poll();
                if (declared.getTypeArguments().size() <= index) {
                    //System.err.println("INVALID: " + declared);
                    return false;
                }
                current = declared.getTypeArguments().get(index);
            } else
                throw new RuntimeException("current: " + current.getKind());
        }

        // Check if the thing we've located is a wildcard.
        if (current instanceof WildcardType)
            return true;

        return false;
    }

    public @Deprecated void annotate(InternalAnnotation annotation) { 
        
        if (annotation.isRaw())
            annotate(annotation.getAnnotationClass());
        else
            annotateAt(annotation.getLocation(), annotation.getAnnotationClass());
    }
    
    /**
     * Manually adds an annotation to the outer type for this class type. It is
     * primarily useful for program elements that don't have symbols from which
     * annotations could be extracted (like typecasts or object creation
     * expressions).
     * 
     * @param annotation
     *            the class token for the annotation type
     */
    public void annotate(Class<? extends Annotation> annotation) {

        if (annotation == null)
            throw new IllegalArgumentException("can't annotate type with null");

        this.outer.add(annotation);
    }

    /**
     * Manually adds an annotation to the inner type for this class type at the
     * specified location. It is primarily useful for program elements that
     * don't have symbols from which annotations could be extracted (like object
     * creation expressions).
     * 
     * @param location
     *            the location at which to add the annotation
     * @param annotation
     *            the class token for the annotation type
     */
    public void annotateAt(AnnotationLocation location,
            Class<? extends Annotation> annotation) {

        // Initialize that location with an empty set if nothing is there.
        if (!this.inner.containsKey(location))
            this.inner
                    .put(location, new HashSet<Class<? extends Annotation>>());

        this.inner.get(location).add(annotation);
    }

    /**
     * Forces an annotation of the given type to be included among the
     * annotations checked by {@code isAnnotatedWith}. This is useful for
     * specifying implicit annotations.
     *
     * @param annotation
     *            the annotation to include among this type's annotations
     */
    public void include(Class<? extends Annotation> annotation) {

        if (annotation == null)
            throw new IllegalArgumentException("can't include null");

        this.includes.add(annotation);
    }

    /**
     * Forces an annotation of the given type to be skipped by
     * {@code isAnnotatedWith}.
     * 
     * @param annotation
     *            the annotation to exclude from this type's annotations
     */
    public void exclude(Class<? extends Annotation> annotation) {

        if (annotation == null)
            throw new IllegalArgumentException("can't exclude null");

        this.excludes.add(annotation);
    }

    /**
     * Sets this type's element reference to the given element.
     *
     * @param e
     *            the element to associate with this type
     */
    void setElement(Element e) {
        this.element = e;
    }

    /**
     * @return the element associated with this type
     */
    public Element getElement() {
        return this.element;
    }
    
    /**
     * Sets this type's tree reference to the given tree.
     *
     * @param t the tree to associate with this type
     */
    void setTree(Tree t) {
        this.tree = t;
    }
    
    /**
     * @return the tree associated with this type
     */
    public Tree getTree() {
        return tree;
    }

    /**
     * @return the locations of all annotations in this class type
     * 
     * @todo should take an annotation argument
     */
    public Set<AnnotationLocation> getAnnotatedLocations() {
        return Collections.unmodifiableSet(this.inner.keySet());
    }

    @Override
    public String toString() {
        return this.outer + " " + this.inner + " " + this.includes + " "
                + this.excludes;
    }
}