package checkers.types;

import java.util.*;

/**
 * Represents the location of an annotation within a type -- either on the raw
 * type, or in the "hierarchy" formed by a generic type. It provides the means
 * to "walk" a generic type to locate the type argument on which an annotation
 * is written.
 */
public final class AnnotationLocation {

    /** The location for an annotation on a raw type. */
    public static final AnnotationLocation RAW = new AnnotationLocation(
            new int[0]);

    /** Element i of this list is the index on the ith level of hierarchy. */
    private final List<Integer> location;

    /**
     * Creates a new instance of {@link AnnotationLocation}.
     * 
     * @param location
     *            the array specifying the location of the annotation, where the
     *            ith array element specifies the index of the type argument on
     *            the ith level of hierarchy
     */
    public AnnotationLocation(int[] location) {

        ArrayList<Integer> lst = new ArrayList<Integer>(location.length);
        for (int i : location)
            lst.add(i);

        this.location = Collections.unmodifiableList(lst);
    }

    /**
     * Creates a new instance of {@link AnnotationLocation}.
     * 
     * @param location
     *            the list specifying the location of the annotation, where the
     *            ith list element specifies the index of the type argument on
     *            the ith level of hierarchy
     */
    public AnnotationLocation(List<Integer> location) {
        this.location = Collections.unmodifiableList(location);
    }

    /**
     * Returns the location as a list of indices in the type hierarchy.
     * 
     * @return the annotation location as a list
     */
    public List<Integer> asList() {
        return Collections.unmodifiableList(location);
    }

    @Override
    public boolean equals(Object o) {
        return (o instanceof AnnotationLocation && ((AnnotationLocation) o).location
                .equals(location));
    }

    @Override
    public int hashCode() {
        return location.hashCode() + 37;
    }
    
    @Override
    public String toString() {
        return location.toString();
    }
}
