package checkers.types;

import java.lang.annotation.Annotation;
import java.util.*;

import javax.lang.model.element.*;

import com.sun.source.tree.Tree;
import com.sun.tools.javac.code.*;
import com.sun.tools.javac.code.Symbol.*;
import com.sun.tools.javac.comp.*;
import com.sun.tools.javac.tree.JCTree;

final class InternalAnnotation {

    /** The annotation's Target. */
    private final AnnotationTarget target;

    /** The annotation's location. */
    private final AnnotationLocation location;

    /** The annotation's class. */
    private final Class<? extends Annotation> annotation;

    /**
     * Creates a new {@link InternalAnnotation}.
     * 
     * @param target
     *            the annotation's target
     * @param location
     *            the annotation's location
     * @param annotation
     *            the annotation's class
     */
    private InternalAnnotation(AnnotationTarget target, AnnotationLocation location,
            Class<? extends Annotation> annotation) {
        this.target = target;
        this.location = location;
        this.annotation = annotation;
    }

    /**
     * Determines whether or not the annotation is an extended annotation.
     * 
     * @return true if the represented annotation is an extended annotation,
     *         false otherwise
     */
    public boolean isExtended() {
        return this.target != null && this.target.type != TargetType.UNKNOWN;
    }

    /**
     * @return the target for this extended annotation
     */
    public AnnotationTarget getTarget() {
        return this.target;
    }

    /**
     * Determines whether or not the annotation is on a raw type.
     * 
     * @return true if the annotation is on a raw type, false if it's on an
     *         inner type
     */
    public boolean isRaw() {
        return !this.target.type.hasLocation();
    }

    /**
     * @return the location of the annotation
     */
    public AnnotationLocation getLocation() {
        return this.location;
    }

    /**
     * @return the extended annotation's class
     */
    public Class<? extends Annotation> getAnnotationClass() {
        return this.annotation;
    }

    /**
     * Produces an {@link ExtendedAnnotation} from an {@link AnnotationMirror}.
     * 
     * @param a
     *            the {@link AnnotationMirror} to use to construct the {@link
     *            ExtendedAnnotation}
     * 
     * @return the {@link ExtendedAnnotation} corresponding to the given {@link
     *         AnnotationMirror}
     */
    @SuppressWarnings("unchecked")
    public static InternalAnnotation fromMirror(AnnotationMirror a) {
        Class<? extends Annotation> clazz;

        try {
            clazz = (Class<Annotation>) Class.forName(a.getAnnotationType()
                    .toString());
        } catch (ClassNotFoundException e) {
            throw new IllegalArgumentException("invalid annotation class");
        }

        if (a instanceof Attribute.Compound) {
            Attribute.Compound attr = (Attribute.Compound) a;
            
            if (attr.target != null) { // if it's extended
        	AnnotationLocation location;

        	if (attr.target.type.hasLocation())
        	    location = new AnnotationLocation(attr.target.location);
        	else
        	    location = AnnotationLocation.RAW;

        	return new InternalAnnotation(attr.target, location, clazz);
            }
        }
        return new InternalAnnotation(null, AnnotationLocation.RAW, clazz);
    }

    /**
     * Creates a list of the annotations inherent in (not manually added to) the
     * given element.
     * 
     * @param element
     *            the element to get annotations for
     * 
     * @return the element's internal annotations
     */
    public static List<InternalAnnotation> fromElement(Element element) {
        List<InternalAnnotation> annotations = new LinkedList<InternalAnnotation>();

        if (element == null)
            return annotations;

        if (!(element instanceof Symbol))
            throw new RuntimeException("element not a symbol");

        Symbol s = (Symbol) element;

        for (AnnotationMirror a : s.attributes_field)
            annotations.add(InternalAnnotation.fromMirror(a));

        return annotations;
    }

    /**
     * Creates a list of the annotations inherent in (not manually added to) the
     * given element that reference the given tree. Useful for annotations that
     * are stored in the body of the enclosing method.
     * 
     * @param e
     *            the source of the annotations
     * @param tree
     *            the referenced tree
     * 
     * @return the annotations on {@code e} that refer to {@code tree}
     */
    public static List<InternalAnnotation> fromElement(Element e,
            Tree tree) {

        List<InternalAnnotation> methodAnnos = fromElement(e);
        List<InternalAnnotation> thisAnnos = new LinkedList<InternalAnnotation>();

        for (InternalAnnotation a : methodAnnos) {
            AnnotationTarget target = a.getTarget();
            if (target == null || target.ref == null)
                continue;
            if (target.ref == tree || target.ref.pos == ((JCTree)tree).pos)
                thisAnnos.add(a);
        }

        return thisAnnos;
    }

    /**
     * For each method parameter in the given element, gathers the annotations
     * on that parameter into a list and returns these lists in a list.
     * 
     * @param element
     *            the method from which parameter annotations will be extracted
     * @return a list of the lists of annotations on each parameter
     */
    public static List<List<InternalAnnotation>> forMethodParameters(
            Element element) {

        List<List<InternalAnnotation>> annotations = new LinkedList<List<InternalAnnotation>>();

        if (element instanceof MethodSymbol)
            for (VarSymbol vs : ((MethodSymbol) element).params())
                annotations.add(fromElement(vs));

        return annotations;
    }

}
