package checkers.types;

import java.util.Collections;

import javax.lang.model.element.*;

import com.sun.source.tree.*;
import com.sun.source.util.*;

/**
 * A scanner for obtaining groups of annotations from a source tree. 
 */
class InternalAnnotationScanner extends TreeScanner<InternalAnnotationGroup, Void> {
    
    /** The syntax tree root. */
    private final CompilationUnitTree root;
    
    /** The Trees instance to use when scanning. */
    private final Trees trees;

    /**
     * Creates a new instance of InternalAnnotationScanner.
     * 
     * @param root the source tree root
     * @param trees the {@link Trees} instance to use for scanning
     */
    public InternalAnnotationScanner(CompilationUnitTree root, Trees trees) {
        this.root = root;
        this.trees = trees;
    }

    @Override
    public InternalAnnotationGroup visitIdentifier(IdentifierTree node, Void p) {
        Element e = InternalUtils.symbol(node);

        return new InternalAnnotationGroup(InternalAnnotation.fromElement(e), e);
    }

    @Override
    public InternalAnnotationGroup visitMemberSelect(MemberSelectTree node,
        Void p) {
        Element e = InternalUtils.symbol(node);

        return new InternalAnnotationGroup(InternalAnnotation.fromElement(e), e);
    }

    @Override
    public InternalAnnotationGroup visitNewClass(NewClassTree node, Void p) {
        assert root != null;

        TreePath path = trees.getPath(root, node);
        Element method = InternalUtils.enclosingSymbol(path);
        // TODO *** this is a hack and needs improvement
        if (method.getKind() == ElementKind.LOCAL_VARIABLE)
            method = method.getEnclosingElement().getEnclosingElement();
        
        return new InternalAnnotationGroup(InternalAnnotation.fromElement(
                method, node), method);
    }

    @Override
    public InternalAnnotationGroup visitTypeCast(TypeCastTree node, Void p) {
        assert root != null;

        TreePath path = trees.getPath(root, node);
        Element method = InternalUtils.symbol(InternalUtils.enclosingMethod(path));

        return new InternalAnnotationGroup(InternalAnnotation.fromElement(
                method, node), method);
    }

    @Override
    public InternalAnnotationGroup visitLiteral(LiteralTree node, Void p) {
        return InternalAnnotationGroup.EMPTY;
    }

    @Override
    public InternalAnnotationGroup visitVariable(VariableTree node, Void p) {
        Element e = InternalUtils.symbol(node);

        return new InternalAnnotationGroup(InternalAnnotation.fromElement(e), e);
    }

    @Override
    public InternalAnnotationGroup visitNewArray(NewArrayTree node, Void p) {
        assert root != null;

        TreePath path = trees.getPath(root, node);
        Element method = InternalUtils.enclosingSymbol(path);

        return new InternalAnnotationGroup(InternalAnnotation.fromElement(
                method, node), method);
    }

    @Override
    public InternalAnnotationGroup visitMethod(MethodTree node, Void p) {
        Element e = InternalUtils.symbol(node);

        return new InternalAnnotationGroup(InternalAnnotation.fromElement(e), e);
    }

    @Override
    public InternalAnnotationGroup visitMethodInvocation(
        MethodInvocationTree node, Void p) {

        Element e = InternalUtils.symbol(node.getMethodSelect());
        TreePath path = TreePath.getPath(root, node);
        return new InternalAnnotationGroup(InternalAnnotation.fromElement(e), e);
    }

    @Override
    public InternalAnnotationGroup visitBinary(BinaryTree node, Void p) {
        return new InternalAnnotationGroup(Collections.<InternalAnnotation>emptyList(), null);
    }
}
